//******************************************************************//
//File Name: generalDriver.c                                        //
//Version:   2.0                                                    //
//Company:   Tri-M Engineering                                      //
//Date:      Oct. 29, 2002                                          //
//                                                                  //
//this file implements the functions declared in generalDriver.h to //
//control following boards                                          //
//                                                                  //
//1. IR104                                                          //
//2. R104-88                                                        //
//3. R104-16DO                                                      //
//4. R104-16DI                                                      //
//5. IO104-60IN                                                     //
//******************************************************************//

#include <stdio.h>
#include <stdlib.h>
#include <sys/io.h>
#include <unistd.h>
#include "generalDriver.h"


//for IR104 board
//******************************************************************
//Name:       IR104SetRelay
//Parameters: addr is base address of IR104 board
//            relay is relay number which you want to set
//Return:     an integer which indicate if this function execute
//            successfully or not
//            return 0   -- sucessful
//            return -1  -- Base address out of range
//            return -2  -- Relay number out of range
//            return -3  -- cannot access the indicated IO address
//Feature:    This function access the given base address and 
//            set the indicated relay and return 0 to indicate 
//            successful operation, if cannot set the relay, then 
//            indicate the cause of failure by corresponding return
//            value  
//******************************************************************
int IR104SetRelay(unsigned int addr, unsigned int relay)
{

  if( addr != 0x240 && addr != 0x260 && addr != 0x280 && addr != 0x300 )
  {//check if the addr is valid base address
    //printf("Error: Base Address Out Of Range.\n"); 
    return -1;
  }
  else if( relay > 20 || relay < 1 )
  {//check if the relay is valid relay number
   // printf("Error: Relay Number Out Of Range.\n");
    return -2;
  }
  else
  {
    unsigned int relayStates;      //relay states before set
    unsigned int newRelay;         //the new relay need to be set
    unsigned int newRelayStates;   //new relay states after set the 
                                   //relay
    
    relayStates = 0;

    //for different relay number, open correspoding IO address for 
    //setting
    if( relay >= 1 && relay <= 8 )    
    {
      if(ioperm(addr, 1, 1) != 0)
      {//open the IO address corresponding to the relay 1-8 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
      
      //get the current relay state
      relayStates = inb(addr);
   
      //get the new relay
      newRelay = 1<<(relay-1);
    
      //set the new relay and at same time keep state of  all other
      //relay
      newRelayStates = relayStates | newRelay;
      outb( newRelayStates, addr);


      if(ioperm(addr, 1, 0) != 0)
      {//close the IO address corresponding to the relay 1-8 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
      
    }
    else if( relay >= 9 && relay <= 16 )
    {
      if(ioperm(addr+1, 1, 1) != 0)
      {//open the IO address corresponding to the relay 9-16 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
      
      //get the current relay state
      relayStates = inb(addr+1);
   
      //get the new relay
      newRelay = 1<<(relay-9);
    
      //set the new relay and at same time keep state of  all other
      //relay
      newRelayStates = relayStates | newRelay;
      outb( newRelayStates, addr+1);


      if(ioperm(addr+1, 1, 0) != 0)
      {//close the IO address corresponding to the relay 9-16 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
      
    }
    else if( relay >= 17 && relay <= 20 )
    {
      if(ioperm(addr+2, 1, 1) != 0)
      {//open the IO address corresponding to the relay 17-20 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
      
      //get the current relay state
      relayStates = inb(addr+2);
   
      //get the new relay
      newRelay = 1<<(relay-17);
    
      //set the new relay and at same time keep state of  all other
      //relay
      newRelayStates = relayStates | newRelay;
      outb( newRelayStates, addr+2);


      if(ioperm(addr+2, 1, 0) != 0)
      {//close the IO address corresponding to the relay 17-20 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
      
    }
    
    //sleep some time and wait for corresponding relay's mechanical 
    //operation 
    usleep(5000);
    
    return 0;
  }
}

//******************************************************************
//Name:       IR104ResetRelay
//Parameters: addr is base address of IR104 board
//            relay is relay number which you want to set
//Return:     an integer which indicate if this function execute
//            successfully or not
//            return 0   -- sucessful
//            return -1  -- Base address out of range
//            return -2  -- Relay number out of range
//            return -3  -- cannot access the indicated IO address
//Feature:    This function access the given base address and 
//            reset the indicated relay and return 0 to indicate 
//            successful operation, if cannot reset the relay, then 
//            indicate the cause of failure by corresponding return
//            value  
//******************************************************************
int IR104ResetRelay(unsigned int addr, unsigned int relay)
{
  if( addr != 0x240 && addr != 0x260 && addr != 0x280 && addr != 0x300 )
  {//check if the addr is valid base address
   // printf("Error: Base Address Out Of Range.\n"); 
    return -1;
  }
  else if(relay > 20 || relay < 1)
  {//check if the relay is valid relay number
   // printf("Error: Relay Number Out Of Range.\n");
    return -2;
  }
  else
  {
    unsigned int relayStates;      //relay states before set
    unsigned int newRelay;         //the relay need to be set
    unsigned int newRelayStates;   //the new relay states after set
   
    //for different relay number, open correspoding IO address for 
    //resetting
    if( relay >= 1 && relay <= 8 )    
    {
      if(ioperm(addr, 1, 1) != 0)
      {//open the IO address corresponding to the relay 1-8 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }

      //get original relay state
      relayStates = inb(addr);
   
      //the new relay state
      newRelay = 1<<(relay-1);
      newRelay = (~newRelay);
    
      //reset the new relay state and keep the original relay state
      newRelayStates = relayStates & newRelay;
      outb(newRelayStates, addr);

      if(ioperm(addr, 1, 0) != 0)
      {//close the IO address corresponding to the relay 1-8 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
      
    }
    else if( relay >= 9 && relay <= 16 )
    {
      if(ioperm(addr+1, 1, 1) != 0)
      {//open the IO address corresponding to the relay 9-16 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }

      //get original relay state
      relayStates = inb(addr+1);
   
      //the new relay state
      newRelay = 1<<(relay-9);
      newRelay = (~newRelay);
    
      //reset the new relay state and keep the original relay state
      newRelayStates = relayStates & newRelay;
      outb(newRelayStates, addr+1);
      
      if(ioperm(addr+1, 1, 0) != 0)
      {//close the IO address corresponding to the relay 9-16 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
    }
    else if( relay >= 17 && relay <= 20 )
    {
      if(ioperm(addr+2, 1, 1) != 0)
      {//open the IO address corresponding to the relay 17-20 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }

      //get original relay state
      relayStates = inb(addr+2);
   
      //the new relay state
      newRelay = 1<<(relay-17);
      newRelay = (~newRelay);
    
      //reset the new relay state and keep the original relay state
      newRelayStates = relayStates & newRelay;
      outb(newRelayStates, addr+2);
      
      if(ioperm(addr+2, 1, 0) != 0)
      {//close the IO address corresponding to the relay 17-20 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
    }

    //waiting for mechanical operation
    usleep(5000);
    
    return 0;
  }  
}

//******************************************************************
//Name:       IR104RelayRead
//Parameters: addr is base address of IR104 board
//            relay is relay number which you want to set
//Return:     an integer which indicate if this function execute
//            successfully or not
//            return 0   -- sucessful
//            return -1  -- Base address out of range
//            return -2  -- Relay number out of range
//            return -3  -- cannot access the indicated IO address
//Feature:    This function access the given base address and 
//            read state of  the indicated relay and return 0 to 
//            indicate successful operation, if cannot set the 
//            relay, then indicate the cause of failure by 
//            corresponding return value  
//******************************************************************
int IR104RelayRead(unsigned int addr, unsigned int relay, unsigned int *value)
{
  if( addr != 0x240 && addr != 0x260 && addr != 0x280 && addr != 0x300 )
  {//check if the addr is valid base address
    //printf("Error: Base Address Out Of Range.\n"); 
    return -1;
  }
  else if(relay > 20 || relay < 1)
  {//check if the relay is valid relay number
    //printf("Error: Relay Number Out Of Range.\n");
    return -2;
  }
  else
  {
    unsigned int relayStates;  //the relay states read out
   
    //for different relay number, open correspoding IO address for 
    //reading
    if( relay >= 1 && relay <= 8 )    
    {
      if(ioperm(addr, 1, 1) != 0)
      {//open the IO address corresponding to the relay 1-8 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
    
      //get relay state
      relayStates = inb(addr);
      //convert relay to value  
      *value = relayStates>>(relay-1) & 1;

      if(ioperm(addr, 1, 0) != 0)
      {//close the IO address corresponding to the relay 1-8 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
    }
    else if( relay >= 9 && relay <= 16 )
    {
      if(ioperm(addr+1, 1, 1) != 0)
      {//open the IO address corresponding to the relay 9-16 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
      
      //get relay state
      relayStates = inb(addr+1);
      //convert relay to value  
      *value = relayStates>>(relay-9) & 1;

      if(ioperm(addr+1, 1, 0) != 0)
      {//close the IO address corresponding to the relay 9-16 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
    }
    else if( relay >= 17 && relay <= 20 )
    {
      if(ioperm(addr+2, 1, 1) != 0)
      {//open the IO address corresponding to the relay 17-20 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
      
      //get relay state
      relayStates = inb(addr+2);
      //convert relay to value  
      *value = relayStates>>(relay-17) & 1;

      if(ioperm(addr+2, 1, 0) != 0)
      {//close the IO address corresponding to the relay 17-20 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
    }
    return 0;
  }
}

//******************************************************************
//Name:       IR104OptoRead
//Parameters: addr    --  base address of IR104 board
//            relay   --  relay number which you want to set
//            *value  --  pointer to opto value(0 or 1)
//Return:     an integer which indicate if this function execute
//            successfully or not
//            return 0   -- sucessful
//            return -1  -- Base address out of range
//            return -2  -- Relay number out of range
//            return -3  -- cannot access the indicated IO address
//Feature:    This function access the given base address and 
//            read  state of the indicated opto input  and return 0 
//            to indicate successful operation, if cannot set the 
//            relay, then indicate the cause of failure by 
//            corresponding return value  
//******************************************************************
int IR104OptoRead(unsigned int addr, unsigned int opto, unsigned int *value)
{

  if( addr != 0x240 && addr != 0x260 && addr != 0x280 && addr != 0x300 )
  {//check if the addr is valid base address
    //printf("Error: Base Address Out Of Range.\n"); 
    return -1;
  }
  else if(opto > 20 || opto < 1)
  {//check if the opto is valid opto input number
    //printf("Error: Opto Number Out Of Range.\n");
    return -2;
  }
  else
  {
    unsigned int optoStates;   //the opto input state read out
  
    //get the opto input address from base address
    addr += 4;
    
    //for different relay number, open correspoding IO address for 
    //reading
    if( opto >= 1 && opto <= 8 )    
    {
      if(ioperm(addr, 1, 1) != 0)
      {//open the IO address corresponding to the relay 1-8 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
    
      //get the relay state and convert to corresponding value
      optoStates = inb(addr);  
      *value = optoStates>>(opto-1) & 1;
    
      if(ioperm(addr, 1, 0) != 0)
      {//close the IO address corresponding to the relay 1-8 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
    }
    else if( opto >= 9 && opto <= 16 )
    {
      if(ioperm(addr+1, 1, 1) != 0)
      {//open the IO address corresponding to the relay 9-16 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
    
    
      //get the relay state and convert to corresponding value
      optoStates = inb(addr+1);  
      *value = optoStates>>(opto-9) & 1;
    
      if(ioperm(addr+1, 1, 0) != 0)
      {//close the IO address corresponding to the relay 9-16 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
    }
    else if( opto >= 17 && opto <= 20 )
    {
      if(ioperm(addr+2, 1, 1) != 0)
      {//open the IO address corresponding to the relay 17-20 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
    
      //get the relay state and convert to corresponding value
      optoStates = inb(addr+2);  
      *value = optoStates>>(opto-17) & 1;
    
      if(ioperm(addr+2, 1, 0) != 0)
      {//close the IO address corresponding to the relay 17-20 for read 
       //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
        return -3;
      }
    }
 
    return 0;
  }
}



//for R104-88 board
unsigned int origRelayStates = 0; //varable to record the origenal
                                  //relay state
//******************************************************************
//Name:       R104_88SetRelay
//Parameters: addr  --  base address of R104-88 board
//            relay --  relay number which you want to set
//Return:     an integer which indicate if this function execute
//            successfully or not
//            return 0   -- sucessful
//            return -1  -- Base address out of range
//            return -2  -- Relay number out of range
//            return -3  -- cannot access the indicated IO address
//Feature:    This function access the given base address and 
//            set the indicated relay and return 0 to indicate 
//            successful operation, if cannot set the relay, then 
//            indicate the cause of failure by corresponding return
//            value  
//******************************************************************
int R104_88SetRelay(unsigned int addr, unsigned int relay)
{
  if( addr != 0x240 && addr != 0x260 && addr != 0x280 && addr != 0x300 )
  {//check if the addr is valid base address
    //printf("Error: Base Address Out Of Range.\n"); 
    return -1;
  }
  else if(relay > 8 || relay < 1)
  {//check if the relay is valid relay number
    //printf("Error: Relay Number Out Of Range.\n");
    return -2;
  }
  else
  {
     unsigned int newRelayStates = 0;  //new relay states after setting
                                       //the indicated one

     if( 1 <= relay && relay <= 4 )
     {
       if(ioperm(addr+2, 1, 1) != 0)
       {//open the IO address corresponding to the relay 1-4 for read 
        //and write 
         // printf("Error: Cannot Access The Address %04X.\n", addr);
         return -3;
       }
      
       //get the new relay state
       newRelayStates = 1<<(relay - 1);
       
       //set only the new relay
       origRelayStates = origRelayStates | newRelayStates;
       outb(origRelayStates, addr+2);

       if(ioperm(addr+2, 1, 0) != 0)
       {//close the IO address corresponding to the relay 1-4 for read 
       //and write 
	// printf("Error: Cannot Access The Address %04X.\n", addr);
         return -3;
       }
     }
     else if( 5 <= relay && relay <= 8)
     {
       if(ioperm(addr+3, 1, 1) != 0)
       {//open the IO address corresponding to the relay 5-8 for read 
        //and write 
        // printf("Error: Cannot Access The Address %04X.\n", addr);
         return -3;
       }
       
       //get the new relay state
       newRelayStates = 1<<(relay - 5);
       
       //set only the new relay
       origRelayStates = origRelayStates | newRelayStates;
       outb(origRelayStates, addr+3);
       
       if(ioperm(addr+3, 1, 0) != 0)
       {//close the IO address corresponding to the relay 5-8 for read 
       //and write 
         // printf("Error: Cannot Access The Address %04X.\n", addr);
         return -3;
       }
     } 

     usleep(5000);
      
     return 0;
   }
}

//******************************************************************
//Name:       R104_88ResetRelay
//Parameters: addr  --  base address of IR104 board
//            relay --  relay number which you want to set
//Return:     an integer which indicate if this function execute
//            successfully or not
//            return 0   -- sucessful
//            return -1  -- Base address out of range
//            return -2  -- Relay number out of range
//            return -3  -- cannot access the indicated IO address
//Feature:    This function access the given base address and 
//            reset the indicated relay and return 0 to indicate 
//            successful operation, if cannot reset the relay, then 
//            indicate the cause of failure by corresponding return
//            value  
//******************************************************************
int R104_88ResetRelay(unsigned int addr, unsigned int relay)
{
   
  if( addr != 0x240 && addr != 0x260 && addr != 0x280 && addr !=  0x300 )
  {//check if the addr is valid base address
    //printf("Error: Base Address Out Of Range.\n"); 
    return -1;
  }
  else if(relay > 8 || relay < 1)
  {//check if the relay is valid relay number
    //printf("Error: Relay Number Out Of Range.\n");
    return -2;
  }
  else
  {
      unsigned int newRelay = 0x0;  //the new relay need been set
      unsigned int newRelayStates;  //the overall new relay states

      if( 1<= relay && relay <= 4)
      {  
         if(ioperm(addr+2, 1, 1) != 0)
         {//open the IO address corresponding to the relay 1-4 for read 
         //and write 
           // printf("Error: Cannot Access The Address %04X.\n", addr);
           return -3;
         }
         
         //get the new relay	 
	 newRelay = 1<<(relay - 1);
	 newRelay = (~newRelay);
         
	 // reset the new relay
         origRelayStates &= newRelay;
         newRelayStates = origRelayStates;
         outb(origRelayStates, addr+2);
	 
	 if(ioperm(addr+2, 1, 0) != 0)
         {//close the IO address corresponding to the relay 1-4 for read 
         //and write 
           // printf("Error: Cannot Access The Address %04X.\n", addr);
           return -3;
         }
      }
      else if( 5 <= relay && relay <= 8 )
      {
         if(ioperm(addr+3, 1, 1) != 0)
         {//open the IO address corresponding to the relay 5-8 for read 
         //and write 
           // printf("Error: Cannot Access The Address %04X.\n", addr);
           return -3;
         }
	 
	 //get the new relay
         newRelay = 1<<(relay - 5);
	 newRelay = (~newRelay);
	 
	 //reset the new relay
         origRelayStates &= newRelay;
         newRelayStates = origRelayStates;
         outb(origRelayStates, addr+3);
         
	 if(ioperm(addr+2, 1, 0) != 0)
         {//close the IO address corresponding to the relay 5-8 for read 
         //and write 
           // printf("Error: Cannot Access The Address %04X.\n", addr);
           return -3;
         }
      }

      usleep(5000);

      return 0;
   }
}

//******************************************************************
//Name:       R104_88OptoRead
//Parameters: addr  --  base address of IR104 board
//            relay --  relay number which you want to set
//            *value  --  pointer to opto value(0 or 1)
//Return:     an integer which indicate if this function execute
//            successfully or not
//            return 0   -- sucessful
//            return -1  -- Base address out of range
//            return -2  -- Relay number out of range
//            return -3  -- cannot access the indicated IO address
//Feature:    This function access the given base address and 
//            the indicated opto, read the state of this opto input
//            and return 0 to indicate successful operation, if 
//            cannot set the relay, then indicate the cause of 
//            failure by corresponding return value  
//******************************************************************
int R104_88OptoRead(unsigned int addr, unsigned int opto, unsigned int *value)
{
  
  if( addr != 0x240 && addr != 0x260 && addr != 0x280 && addr != 0x300 )
  {//check if the addr is valid base address
    //printf("Error: Base Address Out Of Range.\n"); 
    return -1;
  }
  else if(opto > 8 || opto < 1)
  {//check if the opto is valid opto number
    //printf("Error: Opto Number Out Of Range.\n");
    return -2;
  }
  else
  {
      unsigned int optoStates;    //the opto input state read out

      if( 1 <= opto && opto <= 4)
      {
         if(ioperm(addr, 1, 1) != 0)
         {
           // printf("Error: Cannot Access The Address %04X.\n", addr);
           return -3;
         }
         
	 //read the current opto input  state
	 optoStates = inb(addr);
	 
	 //convert the opto states to the opto input value for
	 //the indicated opto value
	 *value = optoStates>>(opto-1) & 1;
      
         if(ioperm(addr, 1, 0) != 0)
         {
           // printf("Error: Cannot Access The Address %04X.\n", addr);
           return -3;
         }
      }
      else if( 5 <= opto && opto <= 8)
      {
         if(ioperm(addr+1, 1, 1) != 0)
         {
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
         }
	 
	 //read the current opto input  state
         optoStates = inb(addr+1);
	 
	 //convert the opto states to the opto input value for
	 //the indicated opto value
	 *value = optoStates>>(opto-5) & 1;
      
         if(ioperm(addr+1, 1, 0) != 0)
         {
           // printf("Error: Cannot Access The Address %04X.\n", addr);
           return -3;
         }
      }
      return 0;
  }
}

//for R104-16DO board
unsigned int origRelayStates2 = 0;
//******************************************************************
//Name:       R104_16DOSetRelay
//Parameters: addr  --  base address of R104_16DO board
//            relay --  relay number which you want to set
//Return:     an integer which indicate if this function execute
//            successfully or not
//            return 0   -- sucessful
//            return -1  -- Base address out of range
//            return -2  -- Relay number out of range
//            return -3  -- cannot access the indicated IO address
//Feature:    This function access the given base address and 
//            set the indicated relay and return 0 to indicate 
//            successful operation, if cannot set the relay, then 
//            indicate the cause of failure by corresponding return
//            value  
//******************************************************************

int R104_16DOSetRelay(unsigned int addr, unsigned int relay)
{

  if( addr != 0x240 && addr != 0x260 && addr != 0x280 && addr != 0x300 )
  {//check if the addr is valid base address
    //printf("Error: Base Address Out Of Range.\n"); 
    return -1;
  }
  else if(relay > 16 || relay < 1)
  {//check if the relay is valid relay number
    //printf("Error: Relay Number Out Of Range.\n");
    return -2;
  }
  else
  {
      unsigned int newRelayStates = 0;   //the new relay need to be set

      if( 1<= relay && relay <= 4 )
      {
        if(ioperm(addr, 1, 1) != 0)
        {
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      
	//get the new relay states
        newRelayStates = 1<<(relay -1);

	//set the new relay states
	origRelayStates2 = origRelayStates2 | newRelayStates;
	outb(origRelayStates2, addr);

        if(ioperm(addr, 1, 0) != 0)
        {
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
	
      }
      else if( 5 <= relay && relay <= 8 )
      {
        if(ioperm(addr+1, 1, 1) != 0)
        {
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      
	//get the new relay state
        newRelayStates = 1<<(relay - 5);

	//set the new relay states
	origRelayStates2 = origRelayStates2 | newRelayStates;
	outb(origRelayStates2, addr+1);

	
        if(ioperm(addr+1, 1, 0) != 0)
        {
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      } 
      else if( 9 <= relay && relay <= 12 )
      {
        if(ioperm(addr+2, 1, 1) != 0)
        {
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      
	//get the new relay state
        newRelayStates = 1<<(relay - 9);

	//set the new relay states
	origRelayStates2 = origRelayStates2 | newRelayStates;
	outb(origRelayStates2, addr+2);

	
        if(ioperm(addr+2, 1, 0) != 0)
        {
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      }
      else if( 13 <= relay && relay <= 16)
      {
        if(ioperm(addr+3, 1, 1) != 0)
        {
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      
	//get the new relay state
        newRelayStates = 1<<(relay - 13);

	//set the new relay states
	origRelayStates2 = origRelayStates2 | newRelayStates;
	outb(origRelayStates2, addr+3);

        if(ioperm(addr+3, 1, 0) != 0)
        {
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
	
      } 
      
      usleep(5000);

      return 0;
  }
}

//******************************************************************
//Name:       R104_16DOResetRelay
//Parameters: addr  --  base address of R104-16DO board
//            relay --  relay number which you want to set
//Return:     an integer which indicate if this function execute
//            successfully or not
//            return 0   -- sucessful
//            return -1  -- Base address out of range
//            return -2  -- Relay number out of range
//            return -3  -- cannot access the indicated IO address
//Feature:    This function access the given base address and 
//            reset the indicated relay and return 0 to indicate 
//            successful operation, if cannot reset the relay, then 
//            indicate the cause of failure by corresponding return
//            value  
//******************************************************************
int R104_16DOResetRelay(unsigned int addr, unsigned int relay)
{

  if( addr != 0x240 && addr != 0x260 && addr != 0x280 && addr != 0x300 )
  {//check if the addr is valid base address
    //printf("Error: Base Address Out Of Range.\n"); 
    return -1;
  }
  else if(relay > 16 || relay < 1)
  {//check if the relay is valid relay number
    //printf("Error: Relay Number Out Of Range.\n");
    return -2;
  }
  else
  {
      unsigned int newRelay;         //the new relay need be set
      unsigned int newRelayStates;   //the new relay states

      newRelay = 1<<(relay -1);
      newRelay = (~newRelay);
      
      
      if( 1<= relay && relay <= 4)
      {
        if(ioperm(addr, 1, 1) != 0)
        {//open the IO address for relay number 1-4
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      
	//get the relay states
        newRelay = 1<<(relay -1);
        newRelay =(~newRelay);
        
	//set the new relay states
        origRelayStates2 &= newRelay;
        newRelayStates = origRelayStates2;
        outb(newRelayStates, addr);
	
	if(ioperm(addr, 1, 0) != 0)
        {//close the IO address for relay number 1-4
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      
      }
      else if( 5 <= relay && relay <= 8)
      {
        if(ioperm(addr+1, 1, 1) != 0)
        {//open the IO address for relay number 5-8
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      
	//get the relay state
        newRelay = 1<<(relay -5);
	newRelay = (~newRelay);
        
	//set the new relay states
        origRelayStates2 &= newRelay;
        newRelayStates = origRelayStates2;
        outb(newRelayStates, addr+1);
	
	if(ioperm(addr+1, 1, 0) != 0)
        {//close the IO address for relay number 5-8
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      
      } 
      else if( 9 <= relay && relay <= 12)
      {
        if(ioperm(addr+2, 1, 1) != 0)
        {//open the IO address for relay number 9-12
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      
	//get the relay state
        newRelay = 1<<(relay -9);
	newRelay = (~newRelay);
        
	//set the new relay states
        origRelayStates2 &= newRelay;
        newRelayStates = origRelayStates2;
        outb(newRelayStates, addr+2);
	
	if(ioperm(addr+2, 1, 0) != 0)
        {//close the IO address for relay number 9-12
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      
      }
      else if( 13 <= relay && relay <= 16)
      {
        if(ioperm(addr+3, 1, 1) != 0)
        {//open the IO address for relay number 13-16
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      
	//get the new relay state
        newRelay = 1<<(relay -13);
	newRelay = (~newRelay);
        
	//set the new relay ststes
        origRelayStates2 &= newRelay;
        newRelayStates = origRelayStates2;
        outb(newRelayStates, addr+3);
	
	if(ioperm(addr+3, 1, 0) != 0)
        {//close the IO address for relay number 13-16
          // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      
      } 

      usleep(5000);

      return 0;
   }
}

//for R104-16DI board
//******************************************************************
//Name:       R104_16DIOptoRead
//Parameters: addr    --  base address of R104-16DI board
//            relay   --  relay number which you want to set
//            *value  --  pointer to opto value(0 or 1)
//Return:     an integer which indicate if this function execute
//            successfully or not
//            return 0   -- sucessful
//            return -1  -- Base address out of range
//            return -2  -- Relay number out of range
//            return -3  -- cannot access the indicated IO address
//Feature:    This function access the given base address and 
//            read  state of the indicated opto input, put the state
//            in this opto input into address of value point to
//            and return 0  to indicate successful operation, if 
//            cannot set the relay, then indicate the cause of 
//            failure by corresponding return value  
//******************************************************************
int R104_16DIOptoRead(unsigned int addr, unsigned int opto, unsigned int *value)
{

  if( addr != 0x240 && addr != 0x260 && addr != 0x280 && addr != 0x300 )
  {//check if the addr is valid base address
    //printf("Error: Base Address Out Of Range.\n"); 
    return -1;
  }
  else if(opto > 16 || opto < 1)
  {//check if the opto is valid opto input number
    //printf("Error: Opto Number Out Of Range.\n");
    return -2;
  }
  else
  {
      unsigned int optoStates;   //the relay states read out
           
      if( 1 <= opto && opto <= 4)
      {
        if(ioperm(addr, 1, 1) != 0)
        {//open the IO address for reading the opto 1-4
         // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        } 
	
	//get the current the opto state
        optoStates = inb(addr);
	//convert opto state to its value
	*value =(optoStates>>(opto - 1)) & 1;     
        
	if(ioperm(addr, 1, 0) != 0)
        {//close the IO address for opto number 1-4
         // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      }
      else if( 5 <= opto && opto <= 8)
      {
        if(ioperm(addr+1, 1, 1) != 0)
        {//open the IO address for reading the opto number 5-8
         // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
	
	//get the current the opto state
        optoStates = inb(addr+1);
	//convert opto state to its value
        *value =(optoStates>>(opto - 5 )) & 1;
        
	if(ioperm(addr+1, 1, 0) != 0)
        {//close the IO address for opto number 5-8
         // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      } 
      else if( 9 <= opto && opto <= 12)
      {
        if(ioperm(addr+2, 1, 1) != 0)
        {//open the IO address for reading the opto number 9-12
         // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
	
	//get the current the opto state
        optoStates = inb(addr+2);
	//convert opto state to its value
	*value =(optoStates>>(opto - 9)) & 1;
        
	if(ioperm(addr+2, 1, 0) != 0)
        {//close the IO address for the opto number 9-12
         // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      }
      else if( 13 <= opto && opto <= 16)
      {
        if(ioperm(addr+3, 1, 1) != 0)
        {//open the IO address for reading the opto 13-16
         // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
        
	//get the current the opto state
	optoStates = inb(addr+3);
	//convert opto state to its value
	*value =(optoStates>>(opto - 13)) & 1;
        
	if(ioperm(addr+3, 1, 0) != 0)
        {//close the IO address for the opto number 13-16
         // printf("Error: Cannot Access The Address %04X.\n", addr);
          return -3;
        }
      }
      return 0;
  }
}


//for IO104-60IN board
//******************************************************************
//Name:       IO104_60INOptoRead
//Parameters: addr    --  base address of IO104_60IN board
//            relay   --  relay number which you want to set
//            *value  --  pointer to opto value(0 or 1)
//Return:     an integer which indicate if this function execute
//            successfully or not
//            return 0   -- sucessful
//            return -1  -- Base address out of range
//            return -2  -- Relay number out of range
//            return -3  -- cannot access the indicated IO address
//Feature:    This function access the given base address and 
//            read  state of the indicated opto input, put the state
//            in this opto input into address of value point to
//            and return 0  to indicate successful operation, if 
//            cannot set the relay, then indicate the cause of 
//            failure by corresponding return value  
//******************************************************************
int IO104_60INOptoRead(unsigned int addr, unsigned int opto, unsigned int * value)
{

  if( addr != 0x240 && addr != 0x260 && addr != 0x280 && addr != 0x300 )
  {//check if the addr is valid base address
    //printf("Error: Base Address Out Of Range.\n"); 
    return -1;
  }
  else if(opto > 60 || opto < 1)
  {//check if the opto is valid opto input number
    //printf("Error: Opto Number Out Of Range.\n");
    return -2;
  }
  else
  {
      unsigned int optoStates;    //the opto input  states read out
  
      if( 1 <= opto && opto <= 8)
      {
         if(ioperm(addr, 1, 1) != 0)
         {//open the IO address addr for reading opto number 1-8
          // printf("Error: Cannot Access The Address %04X.\n", addr);
           return -3;
         }
	 
	 //read the currect opto input state
         optoStates = inb(addr);
	 //convert the opto state to opto value
         *value =(optoStates>>(opto -1)) & 1;     

         if(ioperm(addr, 1, 0) != 0)
         {//close the IO address for opto number 1-8
          // printf("Error: Cannot Access The Address %04X.\n", addr);
           return -3;
         }
      }
      else if( 9 <= opto && opto <= 16 )
      {	 
         if(ioperm(addr+1, 1, 1) != 0)
         {//open the IO address for reading opto number 9-16
            //printf("Error: Cannot Access The Address %04X.\n", addr);
            return -3;
         }

	 //read the currect opto input state
	 optoStates = inb(addr+1);
	 //convert the opto state to opto value
         *value =(optoStates>>(opto-9)) & 1;

         if(ioperm(addr+1, 1, 0) != 0)
         {//close the IO address for opto number 9-16
            //printf("Error: Cannot Access The Address %04X.\n", addr);
            return -3;
         }
      } 
      else if( 17 <= opto && opto <= 24)
      {	 
         if(ioperm(addr+2, 1, 1) != 0)
         {//open the IO address for reading opto number 17-24
            //printf("Error: Cannot Access The Address %04X.\n", addr);
            return -3;
         }
	 
	 //read the currect opto input state
	 optoStates = inb(addr+2);
	 //convert the opto state to opto value
         *value =(optoStates>>(opto-17)) & 1;

         if(ioperm(addr+2, 1, 0) != 0)
         {//close the IO address for opto number 17-24
            //printf("Error: Cannot Access The Address %04X.\n", addr);
            return -3;
         }
      } 
      else if( 25 <= opto && opto <= 32)
      {
         if(ioperm(addr+3, 1, 1) != 0)
         {//open the IO address for reading opto number 25-32
            //printf("Error: Cannot Access The Address %04X.\n", addr);
            return -3;
         }
	 
	 //read the currect opto input state
	 optoStates = inb(addr+3);
	 //convert the opto state to opto value
         *value =(optoStates>>(opto-25)) & 1;

         if(ioperm(addr+3, 1, 0) != 0)
         {//close the IO address for opto number 25-32
            //printf("Error: Cannot Access The Address %04X.\n", addr);
            return -3;
         }
      } 
      else if( 33 <= opto && opto <= 40)
      {	 
         if(ioperm(addr+4, 1, 1) != 0)
         {//open the IO address for reading opto number 33-40
            //printf("Error: Cannot Access The Address %04X.\n", addr);
            return -3;
         }
	 
	 //read the currect opto input state
	 optoStates = inb(addr+4);
	 //convert the opto state to opto value
         *value =(optoStates>>(opto-33)) & 1;

         if(ioperm(addr+4, 1, 0) != 0)
         {//close the IO address for opto number 33-40
            //printf("Error: Cannot Access The Address %04X.\n", addr);
            return -3;
         }
      } 
      else if( 41 <= opto && opto <= 48)
      {	 
         if(ioperm(addr+5, 1, 1) != 0)
         {//open the IO address for reading opto number 41-48
            //printf("Error: Cannot Access The Address %04X.\n", addr);
            return -3;
         }
	 
	 //read the currect opto input state
	 optoStates = inb(addr+5);
	 //convert the opto state to opto value
         *value =(optoStates>>(opto-41)) & 1;

         if(ioperm(addr+5, 1, 0) != 0)
         {//close the IO address for opto number 41-48
            //printf("Error: Cannot Access The Address %04X.\n", addr);
            return -3;
         }
      } 
      else if( 49 <= opto && opto <= 56)
      {	 
         if(ioperm(addr+6, 1, 1) != 0)
         {//open the IO address for reading the opto number 49-56
            //printf("Error: Cannot Access The Address %04X.\n", addr);
            return -3;
         }
	 
	 //read the currect opto input state
	 optoStates = inb(addr+6);
	 //convert the opto state to opto value
         *value =(optoStates>>(opto-49)) & 1;

         if(ioperm(addr+6, 1, 0) != 0)
         {//close the IO address for opto number 49-56
            //printf("Error: Cannot Access The Address %04X.\n", addr);
            return -3;
         }
      } 
      else if( 57 <= opto && opto <= 60)
      {	 
         if(ioperm(addr+7, 1, 1) != 0)
         {//open the IO address for reading the opto number 57-60
            //printf("Error: Cannot Access The Address %04X.\n", addr);
            return -3;
         }
	 
	 //read the currect opto input state
	 optoStates = inb(addr+7);
	 //convert the opto state to opto value
         *value =(optoStates>>(opto-57)) & 1;

         if(ioperm(addr+7, 1, 0) != 0)
         {//close the IO address for the opto number 57-60
            //printf("Error: Cannot Access The Address %04X.\n", addr);
            return -3;
         }
      } 
      return 0;
  }
}

